/* edit-bars.c
 *
 * Copyright 2023 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "crosswords-config.h"
#include <libipuz/libipuz.h>
#include <glib/gi18n-lib.h>
#include "edit-bars.h"
#include "edit-grid.h"
#include "edit-symmetry.h"
#include "xword-state.h"


enum
{
  PROP_0,
  PROP_PUZZLE_STACK,
  PROP_EDIT_GRID,
  PROP_EDIT_SYMMETRY,
  N_PROPS
};

static GParamSpec *obj_props[N_PROPS] =  {NULL, };


struct _EditBars
{
  PanelWidget parent_instance;

  PuzzleStack *puzzle_stack;
  GtkWidget *edit_grid;
  GtkWidget *edit_symmetry;

  GtkWidget *number_label;
  GtkWidget *top_toggle;
  GtkWidget *bottom_toggle;
  GtkWidget *left_toggle;
  GtkWidget *right_toggle;
};


G_DEFINE_FINAL_TYPE (EditBars, edit_bars, PANEL_TYPE_WIDGET);


static void edit_bars_init         (EditBars              *self);
static void edit_bars_class_init   (EditBarsClass         *klass);
static void edit_bars_set_property (GObject               *object,
                                    guint                  prop_id,
                                    const GValue          *value,
                                    GParamSpec            *pspec);
static void edit_bars_get_property (GObject               *object,
                                    guint                  prop_id,
                                    GValue                *value,
                                    GParamSpec            *pspec);
static void edit_bars_dispose      (GObject               *object);
static void barred_toggled_cb      (EditBars              *self,
                                    GtkWidget             *toggle);
static void puzzle_changed_cb      (EditBars              *self,
                                    gboolean               new_frame,
                                    PuzzleStackChangeType  change_type,
                                    PuzzleStack           *puzzle_stack);
static void cursor_changed_cb      (EditBars              *self,
                                    PuzzleStackStage       stage,
                                    IPuzCellCoord         *cursor,
                                    IPuzClueId            *clue_id);


static void
edit_bars_class_init (EditBarsClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->set_property = edit_bars_set_property;
  object_class->get_property = edit_bars_get_property;
  object_class->dispose = edit_bars_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-bars.ui");

  gtk_widget_class_bind_template_child (widget_class, EditBars, number_label);
  gtk_widget_class_bind_template_child (widget_class, EditBars, top_toggle);
  gtk_widget_class_bind_template_child (widget_class, EditBars, bottom_toggle);
  gtk_widget_class_bind_template_child (widget_class, EditBars, left_toggle);
  gtk_widget_class_bind_template_child (widget_class, EditBars, right_toggle);

  gtk_widget_class_bind_template_callback (widget_class, barred_toggled_cb);

  obj_props[PROP_PUZZLE_STACK] = g_param_spec_object ("puzzle-stack",
                                                      NULL, NULL,
                                                      PUZZLE_TYPE_STACK,
                                                      G_PARAM_READWRITE);
  obj_props[PROP_EDIT_GRID] = g_param_spec_object ("edit-grid",
                                                   NULL, NULL,
                                                   EDIT_TYPE_GRID,
                                                   G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
  obj_props[PROP_EDIT_SYMMETRY] = g_param_spec_object ("edit-symmetry",
                                                       NULL, NULL,
                                                       EDIT_TYPE_SYMMETRY,
                                                       G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
  g_object_class_install_properties (object_class, N_PROPS, obj_props);
}

static void
edit_bars_init (EditBars *self)
{
  gtk_widget_init_template (GTK_WIDGET (self));

  gtk_button_set_child (GTK_BUTTON (self->top_toggle), NULL);
  gtk_button_set_child (GTK_BUTTON (self->bottom_toggle), NULL);
  gtk_button_set_child (GTK_BUTTON (self->left_toggle), NULL);
  gtk_button_set_child (GTK_BUTTON (self->right_toggle), NULL);
}

static void
edit_bars_set_property (GObject      *object,
                        guint         prop_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
  EditBars *self = EDIT_BARS (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_clear_object (&self->puzzle_stack);
      self->puzzle_stack = g_value_dup_object (value);
      if (self->puzzle_stack)
        {
          g_signal_connect_swapped (self->puzzle_stack,
                                    "changed",
                                    G_CALLBACK (puzzle_changed_cb),
                                    self);
          puzzle_changed_cb (self, TRUE, STACK_CHANGE_PUZZLE, self->puzzle_stack);
        }
      break;
    case PROP_EDIT_GRID:
      self->edit_grid = g_value_dup_object (value);
      g_signal_connect_swapped (self->edit_grid, "cursor-changed",
                                G_CALLBACK (cursor_changed_cb),
                                self);
      break;
    case PROP_EDIT_SYMMETRY:
      self->edit_symmetry = g_value_dup_object (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_bars_get_property (GObject    *object,
                        guint       prop_id,
                        GValue     *value,
                        GParamSpec *pspec)
{
  EditBars *self = EDIT_BARS (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_value_set_object (value, self->puzzle_stack);
      break;
    case PROP_EDIT_GRID:
      g_value_set_object (value, self->edit_grid);
      break;
    case PROP_EDIT_SYMMETRY:
      g_value_set_object (value, self->edit_symmetry);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_bars_dispose (GObject *object)
{
  EditBars *self;
  GtkWidget *child;

  self = EDIT_BARS (object);

  while ((child = gtk_widget_get_first_child (GTK_WIDGET (object))))
    gtk_widget_unparent (child);

  /* We have refs to these */
  g_clear_object (&self->puzzle_stack);
  g_clear_object (&self->edit_grid);
  g_clear_object (&self->edit_symmetry);

  G_OBJECT_CLASS (edit_bars_parent_class)->dispose (object);

}

static void
barred_toggled_cb (EditBars  *self,
                   GtkWidget *toggle)
{
  IPuzPuzzle *puzzle;
  XwordState *state;
  IPuzCellCoord cursor;
  IPuzClueId clue_id;
  IPuzStyleSides side = 0;
  IPuzStyleSides new_sides = 0;
  g_autoptr (CrosswordsQuirks) quirks;

  if (toggle == self->left_toggle)
    side = IPUZ_STYLE_SIDES_LEFT;
  else if (toggle == self->right_toggle)
    side = IPUZ_STYLE_SIDES_RIGHT;
  else if (toggle == self->top_toggle)
    side = IPUZ_STYLE_SIDES_TOP;
  else if (toggle == self->bottom_toggle)
    side = IPUZ_STYLE_SIDES_BOTTOM;

  puzzle = puzzle_stack_get_puzzle (self->puzzle_stack);
  quirks = crosswords_quirks_new (puzzle);
  crosswords_quirks_set_symmetry (quirks, edit_symmetry_get_symmetry (EDIT_SYMMETRY (self->edit_symmetry)));
  edit_grid_get_cursor (EDIT_GRID (self->edit_grid), &cursor, &clue_id);
  state = xword_state_new (IPUZ_CROSSWORD (puzzle),
                           quirks,
                           XWORD_STATE_EDIT);
  state->cursor = cursor;
  state->clue = clue_id;


  new_sides = ipuz_barred_calculate_side_toggle (IPUZ_BARRED (puzzle),
                                                 cursor, side,
                                                 edit_symmetry_get_symmetry (EDIT_SYMMETRY (self->edit_symmetry)));

  state = xword_state_replace (state,
                               xword_state_set_bars (state, cursor,
                                                     new_sides, TRUE));

  puzzle_stack_push_change (self->puzzle_stack, STACK_CHANGE_GRID, IPUZ_PUZZLE (state->xword));
  edit_grid_set_cursor (EDIT_GRID (self->edit_grid), state->cursor, state->clue);

  xword_state_free (state);
}

static void
edit_bars_update_state (EditBars *self)
{
  IPuzPuzzle *puzzle = NULL;
  IPuzStyleSides bars;
  IPuzCellCoord cursor;
  IPuzClueId clue_id;
  IPuzCell *cell;
  g_autofree gchar *cluenum = NULL;

  puzzle = puzzle_stack_get_puzzle (self->puzzle_stack);
  if (puzzle == NULL)
    return;

  gtk_widget_set_sensitive (self->top_toggle, IPUZ_IS_BARRED (puzzle));
  gtk_widget_set_sensitive (self->left_toggle, IPUZ_IS_BARRED (puzzle));
  gtk_widget_set_sensitive (self->bottom_toggle, IPUZ_IS_BARRED (puzzle));
  gtk_widget_set_sensitive (self->right_toggle, IPUZ_IS_BARRED (puzzle));
  if (! IPUZ_IS_BARRED (puzzle))
    {
      gtk_label_set_label (GTK_LABEL (self->number_label), NULL);
      return;
    }

  edit_grid_get_cursor (EDIT_GRID (self->edit_grid), &cursor, &clue_id);
  bars = ipuz_barred_get_cell_bars (IPUZ_BARRED (puzzle), cursor);

  /* set the toggles */
  g_signal_handlers_block_by_func (self->top_toggle, barred_toggled_cb, self);
  g_signal_handlers_block_by_func (self->left_toggle, barred_toggled_cb, self);
  g_signal_handlers_block_by_func (self->bottom_toggle, barred_toggled_cb, self);
  g_signal_handlers_block_by_func (self->right_toggle, barred_toggled_cb, self);

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (self->left_toggle),
                                IPUZ_STYLE_SIDES_HAS_LEFT (bars));
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (self->right_toggle),
                                IPUZ_STYLE_SIDES_HAS_RIGHT (bars));
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (self->top_toggle),
                                IPUZ_STYLE_SIDES_HAS_TOP (bars));
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (self->bottom_toggle),
                                IPUZ_STYLE_SIDES_HAS_BOTTOM (bars));

  g_signal_handlers_unblock_by_func (self->top_toggle, barred_toggled_cb, self);
  g_signal_handlers_unblock_by_func (self->left_toggle, barred_toggled_cb, self);
  g_signal_handlers_unblock_by_func (self->bottom_toggle, barred_toggled_cb, self);
  g_signal_handlers_unblock_by_func (self->right_toggle, barred_toggled_cb, self);

  /* update the label */
  /* FIXME(LayoutConfig): This is a quick-and-dirty hack to
   * approximate a cell. We should provide a way to extract a
   * LayoutConfig from just one cell so that we can use a PlayCell in
   * place of a label here.
   */
  cell = ipuz_crossword_get_cell (IPUZ_CROSSWORD (puzzle), cursor);
  if (ipuz_cell_get_number (cell))
    cluenum = g_strdup_printf ("%d", ipuz_cell_get_number (cell));
  else if (ipuz_cell_get_label (cell))
    cluenum = g_strdup (ipuz_cell_get_label (cell));
  gtk_label_set_label (GTK_LABEL (self->number_label), cluenum);
}

/* Public methods */

static void
puzzle_changed_cb (EditBars              *self,
                   gboolean               new_frame,
                   PuzzleStackChangeType  change_type,
                   PuzzleStack           *puzzle_stack)
{
  if (change_type == STACK_CHANGE_PUZZLE || change_type == STACK_CHANGE_GRID)
    edit_bars_update_state (self);
}

static void
cursor_changed_cb (EditBars         *self,
                   PuzzleStackStage  stage,
                   IPuzCellCoord    *cursor,
                   IPuzClueId       *clue_id)
{
  edit_bars_update_state (self);
}
