/* edit-clue-details.c
 *
 * Copyright 2021 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "crosswords-config.h"
#include <glib/gi18n-lib.h>
#include <libipuz/libipuz.h>
#include <adwaita.h>
#include "crosswords-misc.h"
#include "edit-clue-details.h"
#include "edit-entry-row.h"
#include "edit-grid.h"
#include "play-grid.h"


enum
{
  PROP_0,
  PROP_PUZZLE_STACK,
  PROP_EDIT_GRID,
  N_PROPS
};

static GParamSpec *obj_props[N_PROPS] =  {NULL, };


struct _EditClueDetails
{
  PanelWidget parent_instance;

  PuzzleStack *puzzle_stack;
  GtkWidget *edit_grid;
  XwordState *answer_state;

  /* Template widgets */
  GtkWidget *clue_row;
  GtkWidget *enumeration_row;
  GtkWidget *answer_grid;
};


static void        edit_clue_details_init                (EditClueDetails       *self);
static void        edit_clue_details_class_init          (EditClueDetailsClass  *klass);
static void        edit_clue_details_set_property        (GObject               *object,
                                                          guint                  prop_id,
                                                          const GValue          *value,
                                                          GParamSpec            *pspec);
static void        edit_clue_details_get_property        (GObject               *object,
                                                          guint                  prop_id,
                                                          GValue                *value,
                                                          GParamSpec            *pspec);
static void        edit_clue_details_dispose             (GObject               *object);
static XwordState *edit_clue_details_create_answer_state (EditClueDetails       *self);
static void        answer_grid_do_command_cb             (PlayGrid              *grid,
                                                          XwordCmdKind           kind,
                                                          EditClueDetails       *self);
static void        answer_grid_cell_selected_cb          (PlayGrid              *grid,
                                                          guint                  row,
                                                          guint                  column,
                                                          EditClueDetails       *self);
static gboolean    clue_row_committed_cb                 (EditClueDetails       *self);
static gboolean    enumeration_row_committed_cb          (EditClueDetails       *self);
static void        edit_clue_details_update_states       (EditClueDetails       *self);
static void        puzzle_changed_cb                     (EditClueDetails       *self,
                                                          gboolean               new_frame,
                                                          PuzzleStackChangeType  change_type,
                                                          PuzzleStack           *puzzle_stack);
static void        cursor_changed_cb                     (EditClueDetails       *self,
                                                          PuzzleStackStage       stage,
                                                          IPuzCellCoord         *cursor,
                                                          IPuzClueId            *clue_id);


G_DEFINE_TYPE (EditClueDetails, edit_clue_details, PANEL_TYPE_WIDGET);


static void
edit_clue_details_init (EditClueDetails *self)
{
  gtk_widget_init_template (GTK_WIDGET (self));

  /* This keeps us at the maximum height for a base_size of 20 */
  gtk_widget_set_size_request (self->answer_grid, -1, 66);

}

static void
edit_clue_details_class_init (EditClueDetailsClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->dispose = edit_clue_details_dispose;
  object_class->set_property = edit_clue_details_set_property;
  object_class->get_property = edit_clue_details_get_property;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-clue-details.ui");

  gtk_widget_class_bind_template_child (widget_class, EditClueDetails, answer_grid);
  gtk_widget_class_bind_template_child (widget_class, EditClueDetails, clue_row);
  gtk_widget_class_bind_template_child (widget_class, EditClueDetails, enumeration_row);
  gtk_widget_class_bind_template_callback (widget_class, answer_grid_do_command_cb);
  gtk_widget_class_bind_template_callback (widget_class, answer_grid_cell_selected_cb);
  gtk_widget_class_bind_template_callback (widget_class, clue_row_committed_cb);
  gtk_widget_class_bind_template_callback (widget_class, enumeration_row_committed_cb);

  gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BIN_LAYOUT);

  obj_props[PROP_PUZZLE_STACK] = g_param_spec_object ("puzzle-stack",
                                                      NULL, NULL,
                                                      PUZZLE_TYPE_STACK,
                                                      G_PARAM_READWRITE);
  obj_props[PROP_EDIT_GRID] = g_param_spec_object ("edit-grid",
                                                   NULL, NULL,
                                                   EDIT_TYPE_GRID,
                                                   G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
  g_object_class_install_properties (object_class, N_PROPS, obj_props);
}

static void
edit_clue_details_set_property (GObject      *object,
                                guint         prop_id,
                                const GValue *value,
                                GParamSpec   *pspec)
{
  EditClueDetails *self = EDIT_CLUE_DETAILS (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_clear_object (&self->puzzle_stack);
      self->puzzle_stack = g_value_dup_object (value);
      if (self->puzzle_stack)
        {
          g_signal_connect_swapped (self->puzzle_stack,
                                    "changed",
                                    G_CALLBACK (puzzle_changed_cb),
                                    self);
          puzzle_changed_cb (self, TRUE, STACK_CHANGE_PUZZLE, self->puzzle_stack);
        }
      break;
    case PROP_EDIT_GRID:
      self->edit_grid = g_value_dup_object (value);
      g_signal_connect_swapped (self->edit_grid, "cursor-changed",
                                G_CALLBACK (cursor_changed_cb),
                                self);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_clue_details_get_property (GObject    *object,
                                guint       prop_id,
                                GValue     *value,
                                GParamSpec *pspec)
{
  EditClueDetails *self = EDIT_CLUE_DETAILS (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_value_set_object (value, self->puzzle_stack);
      break;
    case PROP_EDIT_GRID:
      g_value_set_object (value, self->edit_grid);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_clue_details_dispose (GObject *object)
{
  EditClueDetails *self;
  GtkWidget *child;

  self = EDIT_CLUE_DETAILS (object);

  while ((child = gtk_widget_get_first_child (GTK_WIDGET (object))))
    gtk_widget_unparent (child);

  g_clear_object (&self->puzzle_stack);
  g_clear_pointer (&self->answer_state, xword_state_free);

  G_OBJECT_CLASS (edit_clue_details_parent_class)->dispose (object);
}

static XwordState *
edit_clue_details_create_answer_state (EditClueDetails *self)
{
  IPuzPuzzle *puzzle;
  IPuzCellCoord cursor;
  IPuzClueId clue_id;
  XwordState *new_state;
  g_autofree gchar *str = NULL;
  const gchar *ptr;
  const gchar *end;
  g_autoptr (IPuzCrossword) xword = NULL;
  guint i = 0;

  puzzle = puzzle_stack_get_puzzle (self->puzzle_stack);
  if (puzzle == NULL)
    return NULL;

  edit_grid_get_cursor (EDIT_GRID (self->edit_grid), &cursor, &clue_id);

  if (clue_id.direction == IPUZ_CLUE_DIRECTION_NONE)
    return NULL;

  str = ipuz_crossword_get_clue_string_by_id (IPUZ_CROSSWORD (puzzle), clue_id);

  if (str == NULL || !str[0])
    return NULL;

  xword = ipuz_crossword_new ();
  ipuz_crossword_set_size (xword, utf8_n_clusters (str), 1);
  ptr = str;
  while (ptr[0])
    {
      IPuzCell *cell;
      g_autofree gchar *cluster = NULL;
      IPuzCellCoord coord = {
        .row = 0,
        .column = i
      };

      cluster = utf8_get_next_cluster (ptr, &end);
      ptr = end;

      /* FIXME (magichars): Centralize this */
      if (g_strcmp0 (cluster, "?") != 0)
        {
          cell = ipuz_crossword_get_cell (xword, coord);
          ipuz_cell_set_solution (cell, cluster);
        }
      i++;
    }

  new_state = xword_state_new (xword, NULL, XWORD_STATE_SELECT);

  return new_state;
}

static void
puzzle_changed_cb (EditClueDetails       *self,
                   gboolean               new_frame,
                   PuzzleStackChangeType  change_type,
                   PuzzleStack           *puzzle_stack)
{
  IPuzPuzzle *puzzle;

  puzzle = puzzle_stack_get_puzzle (puzzle_stack);
  if (puzzle == NULL)
    return;

  self->answer_state = xword_state_replace (self->answer_state, edit_clue_details_create_answer_state (self));
  edit_clue_details_update_states (self);
}

static void
cursor_changed_cb (EditClueDetails  *self,
                   PuzzleStackStage  stage,
                   IPuzCellCoord    *cursor,
                   IPuzClueId       *clue_id)
{
  self->answer_state = xword_state_replace (self->answer_state, edit_clue_details_create_answer_state (self));
  edit_clue_details_update_states (self);
}

static void
answer_grid_do_command_cb (PlayGrid        *grid,
                           XwordCmdKind     kind,
                           EditClueDetails *clues)
{
}

static void
answer_grid_cell_selected_cb (PlayGrid        *grid,
                              guint            row,
                              guint            column,
                              EditClueDetails *clues)
{
}

static gboolean
clue_row_committed_cb (EditClueDetails *self)
{
  IPuzPuzzle *puzzle;
  IPuzClueId clue_id;
  IPuzClue *clue;
  const gchar *new_clue_text = NULL;
  const gchar *old_clue_text = NULL;

  puzzle = puzzle_stack_get_puzzle (self->puzzle_stack);

  /* get the texts */
  new_clue_text = gtk_editable_get_text (GTK_EDITABLE (self->clue_row));
  edit_grid_get_cursor (EDIT_GRID (self->edit_grid), NULL, &clue_id);
  clue = ipuz_crossword_get_clue_by_id (IPUZ_CROSSWORD (puzzle), clue_id);
  old_clue_text = ipuz_clue_get_clue_text (clue);

  if (old_clue_text == NULL)
    old_clue_text = "";

  /* We only push if things have really changed */
  if (g_strcmp0 (old_clue_text, new_clue_text) != 0)
    {
      /* FIXME(text): We aren't really handling entities
       * well. IPuzClue::clue_text is defined to be valid */
      ipuz_clue_set_clue_text (clue, new_clue_text);
      puzzle_stack_push_change (self->puzzle_stack,
                                STACK_CHANGE_STATE,
                                puzzle);
      return TRUE;
    }

  return FALSE;
}

static gboolean
enumeration_row_committed_cb (EditClueDetails *self)
{
  IPuzPuzzle *puzzle;
  g_autoptr (IPuzEnumeration) new_enumeration = NULL;
  const gchar *new_enumeration_text = NULL;
  const IPuzEnumeration *old_enumeration;
  IPuzClueId clue_id;
  IPuzClue *clue = NULL;
  const GArray *cells;

  puzzle = puzzle_stack_get_puzzle (self->puzzle_stack);
  edit_grid_get_cursor (EDIT_GRID (self->edit_grid), NULL, &clue_id);
  if (puzzle && clue_id.direction != IPUZ_CLUE_DIRECTION_NONE)
    clue = ipuz_crossword_get_clue_by_id (IPUZ_CROSSWORD (puzzle), clue_id);
  g_assert (clue != NULL);
  cells = ipuz_clue_get_cells (clue);
  g_assert (cells != NULL);

  new_enumeration_text = gtk_editable_get_text (GTK_EDITABLE (self->enumeration_row));
  new_enumeration = ipuz_enumeration_new (new_enumeration_text, IPUZ_VERBOSITY_STANDARD);
  old_enumeration = ipuz_clue_get_enumeration (clue);

  /* Did the enumeration change. This should only trigger if something
   * weird happened*/
  if (old_enumeration != NULL &&
      !g_strcmp0 (ipuz_enumeration_get_src (old_enumeration), new_enumeration_text) )
    return FALSE;

  if (! ipuz_enumeration_get_has_delim (new_enumeration))
    {
      edit_entry_row_set_error (EDIT_ENTRY_ROW (self->enumeration_row),
                                _("_Enumeration — Invalid enumeration"));
      return FALSE;
    }

  ipuz_clue_set_enumeration (clue, new_enumeration);
  puzzle_stack_push_change (self->puzzle_stack,
                            STACK_CHANGE_GRID,
                            puzzle);
  return TRUE;
}

/* one weirdness is we want people to enter simple html and entities
 * as text — but the &apos; and &quot; entities are a bit too
 * far. PangoMarkup will happily consume text with stray quotes in it
 * without complaining, anyway.
 */
static gchar *
clean_markup_for_editing (const gchar *markup)
{
  GString *str;

  str = g_string_new (markup);

  g_string_replace (str, "&apos;", "'", 0);
  g_string_replace (str, "&quot;", "\"", 0);

  return g_string_free_and_steal (str);
}

static void
edit_clue_details_update_states (EditClueDetails *self)
{
  IPuzPuzzle *puzzle;
  IPuzCellCoord cursor;
  IPuzClueId clue_id;
  IPuzClue *clue = NULL;
  g_autofree gchar *clue_text = NULL;
  const gchar *enumeration_display = NULL;
  gboolean showenumerations = FALSE;

  LayoutConfig answer_layout_config = {
    .border_size = 4,
    .base_size = 20,
  };

  g_assert (EDIT_IS_CLUE_DETAILS (self));

  /* Update the clue */
  puzzle = puzzle_stack_get_puzzle (self->puzzle_stack);
  edit_grid_get_cursor (EDIT_GRID (self->edit_grid), &cursor, &clue_id);

  if (puzzle && clue_id.direction != IPUZ_CLUE_DIRECTION_NONE)
    clue = ipuz_crossword_get_clue_by_id (IPUZ_CROSSWORD (puzzle), clue_id);
  g_assert (clue);

  /* clue text */
  clue_text = clean_markup_for_editing (ipuz_clue_get_clue_text (clue));
  edit_entry_row_set_text (EDIT_ENTRY_ROW (self->clue_row), clue_text);

  /* enumeration */
  g_object_get (puzzle,
                "showenumerations", &showenumerations,
                NULL);

  gtk_widget_set_visible (self->enumeration_row, showenumerations);
  if (showenumerations)
    {
      const GArray *cells;
      g_autoptr (IPuzEnumeration) enumeration = NULL;

      enumeration = ipuz_clue_get_enumeration (clue);
      cells = ipuz_clue_get_cells (clue);
      g_assert (cells);
      g_assert (enumeration);

      enumeration_display = ipuz_enumeration_get_display (enumeration);

      edit_entry_row_set_text (EDIT_ENTRY_ROW (self->enumeration_row), enumeration_display);
      if (ipuz_enumeration_delim_length (enumeration) != (gint) cells->len)
        edit_entry_row_set_warning (EDIT_ENTRY_ROW (self->enumeration_row),
                                    _("_Enumeration — Warning: enumeration is a different length"));
    }

  
  /* answer grid */
  gtk_widget_set_visible (self->answer_grid, self->answer_state != NULL);
  if (self->answer_state)
    play_grid_update_state (PLAY_GRID (self->answer_grid), self->answer_state, answer_layout_config);
}

/* Public functions */

/* If we have a clue half-written, commit it to the stack. This is
 * called during stage changes and when saving to disk.
 *
 * One weirdness is that if there are multiple changes, then they get
 * pushed to the stack serially, which may be unexpected for the user.
 */
void
edit_clue_details_commit_changes (EditClueDetails *self)
{
  g_return_if_fail (EDIT_IS_CLUE_DETAILS (self));

  edit_entry_row_commit (EDIT_ENTRY_ROW (self->clue_row));
  edit_entry_row_commit (EDIT_ENTRY_ROW (self->enumeration_row));
}
