/* edit-clue-list.c
 *
 * Copyright 2021 Jonathan Blandford <jrb@gnome.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "crosswords-config.h"
#include <glib/gi18n-lib.h>
#include "crosswords-misc.h"
#include "edit-clue-list.h"
#include "edit-grid.h"
#include "play-clue-list.h"
#include <libipuz/libipuz.h>


enum
{
  PROP_0,
  PROP_PUZZLE_STACK,
  PROP_EDIT_GRID,
  N_PROPS
};


static GParamSpec *obj_props[N_PROPS] =  {NULL, };


struct _EditClueList
{
  GtkWidget parent_instance;

  PuzzleStack *puzzle_stack;
  GtkWidget *edit_grid;

  GtkWidget *primary_clues;
  GtkWidget *secondary_clues;
  XwordState *state;
};


static void edit_clue_list_init         (EditClueList          *self);
static void edit_clue_list_class_init   (EditClueListClass     *klass);
static void edit_clue_list_set_property (GObject               *object,
                                         guint                  prop_id,
                                         const GValue          *value,
                                         GParamSpec            *pspec);
static void edit_clue_list_get_property (GObject               *object,
                                         guint                  prop_id,
                                         GValue                *value,
                                         GParamSpec            *pspec);
static void edit_clue_list_dispose      (GObject               *object);
static void puzzle_changed_cb           (EditClueList          *self,
                                         gboolean               new_frame,
                                         PuzzleStackChangeType  change_type,
                                         PuzzleStack           *puzzle_stack);
static void cursor_changed_cb           (EditClueList          *self,
                                         IPuzCellCoord         *cursor,
                                         IPuzClueId            *clue_id);
static void clue_selected_cb            (EditClueList          *self,
                                         IPuzClueDirection      direction,
                                         gint                   index,
                                         PlayClueList          *clue_list);

G_DEFINE_TYPE (EditClueList, edit_clue_list, PANEL_TYPE_WIDGET);


static void
edit_clue_list_init (EditClueList *self)
{
  gtk_widget_init_template (GTK_WIDGET (self));
}

static void
edit_clue_list_class_init (EditClueListClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->set_property = edit_clue_list_set_property;
  object_class->get_property = edit_clue_list_get_property;
  object_class->dispose = edit_clue_list_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-clue-list.ui");

  gtk_widget_class_bind_template_child (widget_class, EditClueList, primary_clues);
  gtk_widget_class_bind_template_child (widget_class, EditClueList, secondary_clues);

  gtk_widget_class_bind_template_callback (widget_class, clue_selected_cb);


  obj_props[PROP_PUZZLE_STACK] = g_param_spec_object ("puzzle-stack",
                                                      NULL, NULL,
                                                      PUZZLE_TYPE_STACK,
                                                      G_PARAM_READWRITE);
  obj_props[PROP_EDIT_GRID] = g_param_spec_object ("edit-grid",
                                                   NULL, NULL,
                                                   EDIT_TYPE_GRID,
                                                   G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
  g_object_class_install_properties (object_class, N_PROPS, obj_props);
}

static void
edit_clue_list_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
  EditClueList *self = EDIT_CLUE_LIST (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_clear_object (&self->puzzle_stack);
      self->puzzle_stack = g_value_dup_object (value);
      if (self->puzzle_stack)
        {
          g_signal_connect_swapped (self->puzzle_stack,
                                    "changed",
                                    G_CALLBACK (puzzle_changed_cb),
                                    self);
          puzzle_changed_cb (self, TRUE, STACK_CHANGE_PUZZLE, self->puzzle_stack);
        }
      break;
    case PROP_EDIT_GRID:
      self->edit_grid = g_value_dup_object (value);
      g_signal_connect_swapped (self->edit_grid, "cursor-changed",
                                G_CALLBACK (cursor_changed_cb),
                                self);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_clue_list_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
  EditClueList *self = EDIT_CLUE_LIST (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_value_set_object (value, self->puzzle_stack);
      break;
    case PROP_EDIT_GRID:
      g_value_set_object (value, self->edit_grid);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_clue_list_dispose (GObject *object)
{
  EditClueList *self;
  GtkWidget *child;

  self = EDIT_CLUE_LIST (object);

  while ((child = gtk_widget_get_first_child (GTK_WIDGET (object))))
    gtk_widget_unparent (child);

  g_clear_object (&self->puzzle_stack);
  g_clear_object (&self->edit_grid);

  G_OBJECT_CLASS (edit_clue_list_parent_class)->dispose (object);
}

static void
puzzle_changed_cb (EditClueList          *self,
                   gboolean               new_frame,
                   PuzzleStackChangeType  change_type,
                   PuzzleStack           *puzzle_stack)
{
  IPuzPuzzle *puzzle;

  puzzle = puzzle_stack_get_puzzle (puzzle_stack);

  if (puzzle == NULL)
    return;

  self->state = xword_state_replace (self->state,
                                     xword_state_new (IPUZ_CROSSWORD (puzzle),
                                                      NULL, XWORD_STATE_EDIT_BROWSE));
  play_clue_list_update (PLAY_CLUE_LIST (self->primary_clues), self->state, ZOOM_NORMAL);
  play_clue_list_update (PLAY_CLUE_LIST (self->secondary_clues), self->state, ZOOM_NORMAL);
}

static void
cursor_changed_cb (EditClueList  *self,
                   IPuzCellCoord *cursor,
                   IPuzClueId    *clue_id)
{
  g_assert (EDIT_IS_CLUE_LIST (self));
  /* this callback can be triggered by the edit_grid before we get our
   * puzzle set. It's harmless, but we need to check to make sure we
   * have our stack. */
  if (self->state == NULL)
    return;

  if (ipuz_cell_coord_equal (cursor, &self->state->cursor) &&
      ipuz_clue_id_equal (clue_id, &self->state->clue))
    return;

  self->state->cursor = *cursor;
  self->state->clue = *clue_id;
  
  play_clue_list_update (PLAY_CLUE_LIST (self->primary_clues), self->state, ZOOM_NORMAL);
  play_clue_list_update (PLAY_CLUE_LIST (self->secondary_clues), self->state, ZOOM_NORMAL);
}

static void
clue_selected_cb (EditClueList      *self,
                  IPuzClueDirection  direction,
                  gint               index,
                  PlayClueList      *clue_list)
{
  IPuzClue *clue;
  IPuzClueId clue_id = {
    .direction = direction,
    .index = (guint) index,
  };

  clue = ipuz_crossword_get_clue_by_id (IPUZ_CROSSWORD (self->state->xword), clue_id);
  self->state = xword_state_replace (self->state, xword_state_clue_selected (self->state, clue));
  edit_grid_set_cursor (EDIT_GRID (self->edit_grid), self->state->cursor, self->state->clue);
}
