/* edit-grid.c
 *
 * Copyright 2021 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "crosswords-config.h"
#include <libipuz/libipuz.h>
#include <glib/gi18n-lib.h>
#include "crosswords-enums.h"
#include "edit-grid.h"
#include "edit-symmetry.h"
#include "play-grid.h"


enum
{
  PROP_0,
  PROP_STAGE,
  PROP_PUZZLE_STACK,
  PROP_EDIT_SYMMETRY,
  N_PROPS
};

enum
{
  CURSOR_CHANGED,
  N_SIGNALS
};

static GParamSpec *obj_props[N_PROPS] = {NULL, };
static guint obj_signals[N_SIGNALS] = { 0 };


struct _EditGrid
{
  GtkWidget parent_instance;

  PuzzleStack *puzzle_stack;

  /*FIXME: replace with an enum later. */
  PuzzleStackStage stage;
  XwordState *state;
  GtkWidget *edit_symmetry;
  gchar *data_hint;

  /* Template widgets */
  GtkWidget *grid;
};


static void edit_grid_init                     (EditGrid              *self);
static void edit_grid_class_init               (EditGridClass         *klass);
static void edit_grid_set_property             (GObject               *object,
                                                guint                  prop_id,
                                                const GValue          *value,
                                                GParamSpec            *pspec);
static void edit_grid_get_property             (GObject               *object,
                                                guint                  prop_id,
                                                GValue                *value,
                                                GParamSpec            *pspec);
static void edit_grid_dispose                  (GObject               *object);
static void edit_grid_activate_clipboard_copy  (EditGrid              *self,
                                                const char            *name,
                                                GVariant              *parameter);
static void edit_grid_activate_clipboard_paste (EditGrid              *self,
                                                const char            *name,
                                                GVariant              *parameter);
static void grid_guess_cb                      (PlayGrid              *grid,
                                                gchar                 *guess,
                                                EditGrid              *self);
static void grid_guess_at_cell_cb              (PlayGrid              *grid,
                                                gchar                 *guess,
                                                guint                  row,
                                                guint                  column,
                                                EditGrid              *self);
static void grid_do_command_cb                 (PlayGrid              *grid,
                                                XwordCmdKind           kind,
                                                EditGrid              *xword);
static void grid_cell_selected_cb              (PlayGrid              *grid,
                                                guint                  row,
                                                guint                  column,
                                                EditGrid              *self);
static void edit_grid_update_states            (EditGrid              *self);
static void edit_grid_push_change              (EditGrid              *self,
                                                IPuzCellCoord          old_cursor,
                                                IPuzClueId             old_clue_id);
static void puzzle_changed_cb                  (EditGrid              *self,
                                                gboolean               new_frame,
                                                PuzzleStackChangeType  change_type,
                                                PuzzleStack           *puzzle_stack);


G_DEFINE_TYPE (EditGrid, edit_grid, GTK_TYPE_WIDGET);

/* keep a unique id for every instance of EditGrid so it can find its
 * own puzzle-stack data. */
static guint data_hint_id = 0;

static void
edit_grid_init (EditGrid *self)
{
  self->stage = EDIT_STAGE_GRID;
  gtk_widget_init_template (GTK_WIDGET (self));
  self->data_hint = g_strdup_printf ("edit-grid-%u", data_hint_id++);
}

static void
edit_grid_class_init (EditGridClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->set_property = edit_grid_set_property;
  object_class->get_property = edit_grid_get_property;
  object_class->dispose = edit_grid_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-grid.ui");

  gtk_widget_class_bind_template_callback (widget_class, grid_guess_cb);
  gtk_widget_class_bind_template_callback (widget_class, grid_guess_at_cell_cb);
  gtk_widget_class_bind_template_callback (widget_class, grid_do_command_cb);
  gtk_widget_class_bind_template_callback (widget_class, grid_cell_selected_cb);

  gtk_widget_class_bind_template_child (widget_class, EditGrid, grid);

  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_c, GDK_CONTROL_MASK, "clipboard.copy", NULL);
  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_v, GDK_CONTROL_MASK, "clipboard.paste", NULL);

  gtk_widget_class_install_action (widget_class, "clipboard.copy", NULL,
                                   (GtkWidgetActionActivateFunc) edit_grid_activate_clipboard_copy);
  gtk_widget_class_install_action (widget_class, "clipboard.paste", NULL,
                                   (GtkWidgetActionActivateFunc) edit_grid_activate_clipboard_paste);

  gtk_widget_class_set_css_name (widget_class, "edit-grid");

  obj_signals [CURSOR_CHANGED] =
    g_signal_new ("cursor-changed",
                  EDIT_TYPE_GRID,
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
                  0,
                  NULL, NULL,
                  NULL,
                  G_TYPE_NONE, 2,
                  IPUZ_TYPE_CELL_COORD,
                  IPUZ_TYPE_CLUE_ID);
  
  obj_props[PROP_PUZZLE_STACK] = g_param_spec_object ("puzzle-stack",
                                                      NULL, NULL,
                                                      PUZZLE_TYPE_STACK,
                                                      G_PARAM_READWRITE);
  obj_props[PROP_STAGE] = g_param_spec_enum ("stage", NULL, NULL,
                                             PUZZLE_TYPE_STACK_STAGE,
                                             EDIT_STAGE_GRID,
                                             G_PARAM_READWRITE|G_PARAM_CONSTRUCT_ONLY);
  obj_props[PROP_EDIT_SYMMETRY] = g_param_spec_object ("edit-symmetry",
                                                       NULL, NULL,
                                                       EDIT_TYPE_SYMMETRY,
                                                       G_PARAM_READWRITE);
  g_object_class_install_properties (object_class, N_PROPS, obj_props);

  gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BIN_LAYOUT);
}

static void
edit_grid_set_property (GObject      *object,
                        guint         prop_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
  EditGrid *self;

  self = EDIT_GRID (object);

  switch (prop_id)
    {
    case PROP_STAGE:
      self->stage = g_value_get_enum (value);
      if (self->stage == EDIT_STAGE_GRID)
        gtk_widget_add_css_class (GTK_WIDGET (object), "stage-grid");
      else if (self->stage == EDIT_STAGE_CLUES)
        gtk_widget_add_css_class (GTK_WIDGET (object), "stage-clues");
      break;
    case PROP_PUZZLE_STACK:
      g_clear_object (&self->puzzle_stack);
      self->puzzle_stack = g_value_dup_object (value);
      if (self->puzzle_stack)
        {
          g_signal_connect_swapped (self->puzzle_stack,
                                    "changed",
                                    G_CALLBACK (puzzle_changed_cb),
                                    self);
          puzzle_changed_cb (self, TRUE, STACK_CHANGE_PUZZLE, self->puzzle_stack);
        }
      break;
    case PROP_EDIT_SYMMETRY:
      self->edit_symmetry = g_value_dup_object (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_grid_get_property (GObject    *object,
                        guint       prop_id,
                        GValue     *value,
                        GParamSpec *pspec)
{
  EditGrid *self;

  self = EDIT_GRID (object);

  switch (prop_id)
    {
    case PROP_STAGE:
      g_value_set_enum (value, self->stage);
      break;
    case PROP_PUZZLE_STACK:
      g_value_set_object (value, self->puzzle_stack);
      break;
    case PROP_EDIT_SYMMETRY:
      g_value_set_object (value, self->edit_symmetry);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
edit_grid_dispose (GObject *object)
{
  EditGrid *self;
  GtkWidget *child;

  self = EDIT_GRID (object);

  while ((child = gtk_widget_get_first_child (GTK_WIDGET (object))))
    gtk_widget_unparent (child);

  g_clear_pointer (&self->state, xword_state_free);
  g_clear_object (&self->puzzle_stack);
  g_clear_object (&self->edit_symmetry);
  g_clear_pointer (&self->data_hint, g_free);

  G_OBJECT_CLASS (edit_grid_parent_class)->dispose (object);
}

static void
edit_grid_activate_clipboard_copy (EditGrid   *self,
                                   const char *name,
                                   GVariant   *parameter)
{
  GdkClipboard *clipboard;
  g_autofree gchar *str = NULL;

  if (self->state == NULL)
    return;

  /* If direction is NONE, we're probably on a block */
  if (self->state->clue.direction == IPUZ_CLUE_DIRECTION_NONE)
    return;

  str = ipuz_crossword_get_clue_string_by_id (self->state->xword,
                                              self->state->clue);
  clipboard = gtk_widget_get_clipboard (GTK_WIDGET (self));
  gdk_clipboard_set_text (clipboard, str);
}

static void
paste_received (GObject      *clipboard,
                GAsyncResult *result,
                gpointer      data)
{
  g_autoptr (EditGrid) self = EDIT_GRID (data); /* There's an outstanding ref on self */
  IPuzCellCoord old_cursor = self->state->cursor;
  IPuzClueId old_clue_id = self->state->clue;
  g_autofree gchar *text = NULL;
  g_autofree gchar *upper = NULL;
  IPuzClueDirection direction;

  text = gdk_clipboard_read_text_finish (GDK_CLIPBOARD (clipboard), result, NULL);
  if (text == NULL || self->state == NULL)
    {
      /* Strange text to get. */
      gtk_widget_error_bell (GTK_WIDGET (self));
      return;
    }

  direction = self->state->clue.direction;
  upper = g_utf8_strup (text, -1);
  /* If we're on a BLOCK, arbitrarily paste across. */
  if (direction == IPUZ_CLUE_DIRECTION_NONE)
    direction = IPUZ_CLUE_DIRECTION_ACROSS;

  self->state = xword_state_replace (self->state,
                                     xword_state_guess_word (self->state,
                                                             self->state->cursor,
                                                             direction, upper));

  edit_grid_push_change (self, old_cursor, old_clue_id);
}

static void
edit_grid_activate_clipboard_paste (EditGrid   *self,
                                    const char *name,
                                    GVariant   *parameter)
{
  GdkClipboard *clipboard;

  if (self->state == NULL)
    return;

  clipboard = gtk_widget_get_clipboard (GTK_WIDGET (self));
  gdk_clipboard_read_text_async (clipboard, NULL, paste_received, g_object_ref (self));

}

/* Called when a new frame is pushed on the stack */
static void
puzzle_changed_new_frame (EditGrid    *self,
                          PuzzleStack *puzzle_stack)
{
  IPuzPuzzle *puzzle;
  PuzzleStackChangeType change_type;

  puzzle = puzzle_stack_get_puzzle (puzzle_stack);
  change_type = puzzle_stack_get_change_type (puzzle_stack);

  /* We always create a new state and reset the state when we get a new puzzle */
  if (change_type == STACK_CHANGE_PUZZLE)
    {
      XwordStateMode mode;
      g_autoptr (CrosswordsQuirks) quirks;

      switch (self->stage)
        {
        case EDIT_STAGE_GRID:
          mode = XWORD_STATE_EDIT;
          break;
        case EDIT_STAGE_CLUES:
          mode = XWORD_STATE_EDIT_BROWSE;
          break;
        default:
          g_assert_not_reached ();
        }

      quirks = crosswords_quirks_new (puzzle);
      crosswords_quirks_set_symmetry (quirks, edit_symmetry_get_symmetry (EDIT_SYMMETRY (self->edit_symmetry)));
      self->state = xword_state_replace (self->state, xword_state_new (IPUZ_CROSSWORD (puzzle), quirks, mode));
    }
  else
    {
      /* We may have a similar looking puzzle, but it's a new-to-us puzzle
       * so we update the state as best as we can */
      if (self->state->xword != IPUZ_CROSSWORD (puzzle))
        {
          self->state = xword_state_replace (self->state, xword_state_dehydrate (self->state));
          self->state = xword_state_replace (self->state, xword_state_hydrate (self->state,
                                                                               IPUZ_CROSSWORD (puzzle),
                                                                               NULL));
        }
    }

  if (self->state)
    {
      XwordState *dehydrated_state;
      dehydrated_state = xword_state_dehydrate (self->state);
      puzzle_stack_set_data (puzzle_stack,
                             self->data_hint,
                             dehydrated_state,
                             (GDestroyNotify) xword_state_free);
      g_signal_emit (self, obj_signals [CURSOR_CHANGED], 0, &self->state->cursor, &self->state->clue);
    }
}

static void
puzzle_changed_load_frame (EditGrid    *self,
                           PuzzleStack *puzzle_stack)
{
  IPuzPuzzle *puzzle;
  XwordState *dehydrated_state;

  puzzle = puzzle_stack_get_puzzle (puzzle_stack);
  dehydrated_state = puzzle_stack_get_data (puzzle_stack, self->data_hint);

  if (dehydrated_state)
    {
      g_autoptr (CrosswordsQuirks) quirks;
      IPuzCellCoord cursor = self->state->cursor;
      IPuzClueId clue_id = self->state->clue;

      quirks = crosswords_quirks_new (puzzle);
      crosswords_quirks_set_symmetry (quirks, edit_symmetry_get_symmetry (EDIT_SYMMETRY (self->edit_symmetry)));
      self->state = xword_state_replace (self->state, xword_state_hydrate (dehydrated_state,
                                                                           IPUZ_CROSSWORD (puzzle),
                                                                           quirks));
      if (! ipuz_cell_coord_equal (&cursor, &self->state->cursor) ||
          ! ipuz_clue_id_equal (&clue_id, &self->state->clue))
        g_signal_emit (self, obj_signals [CURSOR_CHANGED], 0, &self->state->cursor, &self->state->clue);
    }
}

static void
puzzle_changed_cb (EditGrid              *self,
                   gboolean               new_frame,
                   PuzzleStackChangeType  change_type,
                   PuzzleStack           *puzzle_stack)
{
  IPuzPuzzle *puzzle;

  puzzle = puzzle_stack_get_puzzle (puzzle_stack);

  if (puzzle == NULL)
    return;

  if (new_frame)
    puzzle_changed_new_frame (self, puzzle_stack);
  else
    puzzle_changed_load_frame (self, puzzle_stack);

  if (change_type == STACK_CHANGE_PUZZLE ||
      change_type == STACK_CHANGE_GRID)
    edit_grid_update_states (self);
}

static void
grid_guess_cb (PlayGrid *grid,
               gchar    *guess,
               EditGrid *xword)
{
  IPuzCellCoord old_cursor = xword->state->cursor;
  IPuzClueId old_clue_id = xword->state->clue;

  crosswords_quirks_set_symmetry (xword->state->quirks, edit_symmetry_get_symmetry (EDIT_SYMMETRY (xword->edit_symmetry)));
  xword->state = xword_state_replace (xword->state, xword_state_guess (xword->state, guess));

  edit_grid_push_change (xword, old_cursor, old_clue_id);
}

static void
grid_guess_at_cell_cb (PlayGrid *grid,
                       gchar    *guess,
                       guint     row,
                       guint     column,
                       EditGrid *self)
{
  IPuzCellCoord old_cursor = self->state->cursor;
  IPuzClueId old_clue_id = self->state->clue;
  IPuzCellCoord coord = {
    .row = row,
    .column = column
  };

  self->state = xword_state_replace (self->state, xword_state_guess_at_cell (self->state, guess, coord));

  edit_grid_push_change (self, old_cursor, old_clue_id);
}


static void
grid_do_command_cb (PlayGrid     *grid,
                    XwordCmdKind  kind,
                    EditGrid     *self)
{
  IPuzCellCoord old_cursor = self->state->cursor;
  IPuzClueId old_clue_id = self->state->clue;

  self->state = xword_state_replace (self->state, xword_state_do_command (self->state, kind));

  edit_grid_update_states (self);

  if (XWORD_KIND_DESTRUCTIVE (kind))
    {
      /* FIXME(undo): By pushing this change, we are growing the stack
       * regardless of whether or not we deleted something */
      edit_grid_push_change (self, old_cursor, old_clue_id);
    }
  else
    {
      /* no need to push a change for a cursor move. Just update the
       * grid and potentially notify a cursor change */
      edit_grid_update_states (self);
      g_signal_emit (self, obj_signals [CURSOR_CHANGED], 0, &self->state->cursor, &self->state->clue);
    }
}

static void
grid_cell_selected_cb (PlayGrid *grid,
                       guint     row,
                       guint     column,
                       EditGrid *self)
{
  IPuzCellCoord coord = {
    .row = row,
    .column = column,
  };
  self->state = xword_state_replace (self->state, xword_state_cell_selected (self->state, coord));

  edit_grid_update_states (self);
  g_signal_emit (self, obj_signals [CURSOR_CHANGED], 0, &self->state->cursor, &self->state->clue);
}

static void
edit_grid_update_states (EditGrid *self)
{
  if (self->state == NULL)
    return;

  play_grid_update_state (PLAY_GRID (self->grid),
                          self->state,
                          layout_config_default (IPUZ_PUZZLE_CROSSWORD));

  /* We get updated when the stack changes even when we're not being
   * shown. Grabbing focus can confuse gtk. Only set the focus when
   * we're mapped. */
  if (gtk_widget_get_mapped (GTK_WIDGET (self)) &&
      XWORD_STATE_HAS_CURSOR (self->state))
    play_grid_focus_cell (PLAY_GRID (self->grid), self->state->cursor);
}

static void
edit_grid_push_change (EditGrid      *self,
                       IPuzCellCoord  old_cursor,
                       IPuzClueId     old_clue_id)
{
  g_assert (EDIT_IS_GRID (self));

  puzzle_stack_push_change (self->puzzle_stack, STACK_CHANGE_GRID,
                            IPUZ_PUZZLE (self->state->xword));
  g_signal_emit (self, obj_signals [CURSOR_CHANGED], 0, &self->state->cursor, &self->state->clue);
}

/* Public methods */

void
edit_grid_get_cursor (EditGrid      *self,
                      IPuzCellCoord *cursor,
                      IPuzClueId    *clue_id)
{
  g_return_if_fail (EDIT_IS_GRID (self));

  g_assert (self->state);

  if (cursor)
    *cursor = self->state->cursor;
  if (clue_id)
    *clue_id = self->state->clue;
}

void
edit_grid_set_cursor(EditGrid      *self,
                     IPuzCellCoord  cursor,
                     IPuzClueId     clue_id)
{
  g_return_if_fail (EDIT_IS_GRID (self));

  g_assert (self->state);

  if (! ipuz_cell_coord_equal (&cursor, &self->state->cursor) ||
      ! ipuz_clue_id_equal (&clue_id, &self->state->clue))
    {
      self->state->cursor = cursor;
      self->state->clue = clue_id;
      edit_grid_update_states (self);

      g_signal_emit (self, obj_signals [CURSOR_CHANGED], 0, &self->state->cursor, &self->state->clue);
    }
}
