/* edit-symmetry.c
 *
 * Copyright 2023 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "crosswords-config.h"
#include <libipuz/libipuz.h>
#include <glib/gi18n-lib.h>
#include "edit-symmetry.h"


enum
{
  PROP_0,
  PROP_PUZZLE_STACK,
  N_PROPS
};

static GParamSpec *obj_props[N_PROPS] =  {NULL, };


struct _EditSymmetry
{
  PanelWidget parent_instance;

  PuzzleStack *puzzle_stack;
  IPuzSymmetry symmetry;

  GtkWidget *symmetry_icon;
  GtkWidget *symmetry_drop_down;
};


G_DEFINE_FINAL_TYPE (EditSymmetry, edit_symmetry, PANEL_TYPE_WIDGET);


static void edit_symmetry_init         (EditSymmetry          *self);
static void edit_symmetry_class_init   (EditSymmetryClass     *klass);
static void edit_symmetry_set_property (GObject               *object,
                                        guint                  prop_id,
                                        const GValue          *value,
                                        GParamSpec            *pspec);
static void edit_symmetry_get_property (GObject               *object,
                                        guint                  prop_id,
                                        GValue                *value,
                                        GParamSpec            *pspec);
static void edit_symmetry_dispose      (GObject               *object);
static void puzzle_changed_cb          (EditSymmetry          *self,
                                        gboolean               new_frame,
                                        PuzzleStackChangeType  change_type,
                                        PuzzleStack           *puzzle_stack);
static void update_icon                (EditSymmetry          *self);
static void drop_down_changed_cb       (EditSymmetry          *self);


static void
edit_symmetry_class_init (EditSymmetryClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->set_property = edit_symmetry_set_property;
  object_class->get_property = edit_symmetry_get_property;
  object_class->dispose = edit_symmetry_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-symmetry.ui");
  gtk_widget_class_bind_template_child (widget_class, EditSymmetry, symmetry_icon);
  gtk_widget_class_bind_template_child (widget_class, EditSymmetry, symmetry_drop_down);
  gtk_widget_class_bind_template_callback (widget_class, drop_down_changed_cb);

  obj_props[PROP_PUZZLE_STACK] = g_param_spec_object ("puzzle-stack",
                                                      NULL, NULL,
                                                      PUZZLE_TYPE_STACK,
                                                      G_PARAM_READWRITE);
 g_object_class_install_properties (object_class, N_PROPS, obj_props);
}

static void
edit_symmetry_init (EditSymmetry *self)
{
  self->symmetry = IPUZ_SYMMETRY_ROTATIONAL_QUARTER;
  gtk_widget_init_template (GTK_WIDGET (self));
  update_icon (self);
}


static void
edit_symmetry_set_property (GObject      *object,
                            guint         prop_id,
                            const GValue *value,
                            GParamSpec   *pspec)
{
  EditSymmetry *self = EDIT_SYMMETRY (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_clear_object (&self->puzzle_stack);
      self->puzzle_stack = g_value_dup_object (value);
      if (self->puzzle_stack)
        {
          g_signal_connect_swapped (self->puzzle_stack,
                                    "changed",
                                    G_CALLBACK (puzzle_changed_cb),
                                    self);
          puzzle_changed_cb (self, TRUE, STACK_CHANGE_PUZZLE, self->puzzle_stack);
        }
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_symmetry_get_property (GObject    *object,
                            guint       prop_id,
                            GValue     *value,
                            GParamSpec *pspec)
{
  EditSymmetry *self = EDIT_SYMMETRY (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_value_set_object (value, self->puzzle_stack);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_symmetry_dispose (GObject *object)
{
  EditSymmetry *self;
  GtkWidget *child;

  self = EDIT_SYMMETRY (object);

  while ((child = gtk_widget_get_first_child (GTK_WIDGET (object))))
    gtk_widget_unparent (child);

  g_clear_object (&self->puzzle_stack);

  G_OBJECT_CLASS (edit_symmetry_parent_class)->dispose (object);

}

static void
puzzle_changed_cb (EditSymmetry          *self,
                   gboolean               new_frame,
                   PuzzleStackChangeType  change_type,
                   PuzzleStack           *puzzle_stack)
{
  IPuzPuzzle *puzzle;

  puzzle = puzzle_stack_get_puzzle (puzzle_stack);
  if (puzzle == NULL)
    return;

  if (change_type == STACK_CHANGE_PUZZLE)
    {
      /* Match the symmetry to the natural symmetry in the puzzle */
      self->symmetry = ipuz_crossword_get_symmetry (IPUZ_CROSSWORD (puzzle));
      gtk_drop_down_set_selected (GTK_DROP_DOWN (self->symmetry_drop_down), self->symmetry);
    }
}

static void
update_icon (EditSymmetry *self)
{
  const gchar *icon_name = NULL;
  switch (self->symmetry)
    {
    case IPUZ_SYMMETRY_NONE:
      icon_name = "symmetry-none";
      break;
    case IPUZ_SYMMETRY_ROTATIONAL_QUARTER:
      icon_name = "symmetry-quarter";
      break;
    case IPUZ_SYMMETRY_ROTATIONAL_HALF:
      icon_name = "symmetry-half";
      break;
    case IPUZ_SYMMETRY_HORIZONTAL:
      icon_name = "symmetry-horizontal";
      break;
    case IPUZ_SYMMETRY_VERTICAL:
      icon_name = "symmetry-vertical";
      break;
    case IPUZ_SYMMETRY_MIRRORED:
      icon_name = "symmetry-mirrored";
      break;
    default:
      icon_name = "broken";
      break;
    }
  gtk_image_set_from_icon_name (GTK_IMAGE (self->symmetry_icon), icon_name);
}

static void
drop_down_changed_cb (EditSymmetry *self)
{
  self->symmetry = (IPuzSymmetry) gtk_drop_down_get_selected (GTK_DROP_DOWN (self->symmetry_drop_down));
  update_icon (self);
}

/* Public methods */

IPuzSymmetry
edit_symmetry_get_symmetry (EditSymmetry *self)
{
  return self->symmetry;
}
