/* edit-word-list.c
 *
 * Copyright 2023 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "crosswords-config.h"
#include <libipuz/libipuz.h>
#include <glib/gi18n-lib.h>
#include "crosswords-enums.h"
#include "edit-grid.h"
#include "edit-word-list.h"
#include "play-grid.h"
#include "word-list.h"
#include "word-list-model.h"


enum
{
  PROP_0,
  PROP_PUZZLE_STACK,
  PROP_EDIT_GRID,
  N_PROPS
};

static GParamSpec *obj_props[N_PROPS] =  {NULL, };


struct _EditWordList
{
  PanelWidget parent_instance;

  PuzzleStack *puzzle_stack;
  gulong changed_id;
  GtkWidget *edit_grid;

  XwordState *state;
  WordListModel *across_list_model;
  WordListModel *down_list_model;

  GtkWidget *across_word_list_view;
  GtkWidget *down_word_list_view;
};


G_DEFINE_TYPE (EditWordList, edit_word_list, PANEL_TYPE_WIDGET);


static void edit_word_list_init          (EditWordList          *self);
static void edit_word_list_class_init    (EditWordListClass     *klass);
static void edit_word_list_set_property  (GObject               *object,
                                          guint                  prop_id,
                                          const GValue          *value,
                                          GParamSpec            *pspec);
static void edit_word_list_get_property  (GObject               *object,
                                          guint                  prop_id,
                                          GValue                *value,
                                          GParamSpec            *pspec);
static void edit_word_list_dispose       (GObject               *object);
static void across_word_activate_cb      (EditWordList          *self,
                                          guint                  position);
static void down_word_activate_cb        (EditWordList          *self,
                                          guint                  position);
static void edit_word_list_update_state  (EditWordList          *self,
                                          XwordState            *state);
static void edit_word_list_update_states (EditWordList          *self);
static void puzzle_changed_cb            (EditWordList          *self,
                                          gboolean               new_frame,
                                          PuzzleStackChangeType  change_type,
                                          PuzzleStack           *puzzle_stack);
static void grid_cursor_changed_cb       (EditWordList          *self,
                                          IPuzCellCoord         *cursor,
                                          IPuzClueId            *clue_id);


static void
setup_listitem_cb (GtkListItemFactory *factory,
                   GtkListItem        *list_item)
{
  GtkWidget *label;

  label = gtk_label_new (NULL);
  gtk_label_set_xalign (GTK_LABEL (label), 0.0);
  gtk_list_item_set_child (list_item, label);
}

static void
bind_listitem_cb (GtkListItemFactory *factory,
                  GtkListItem        *list_item)
{
  GtkWidget *label;
  WordListModelRow *row;
  g_autofree gchar *text = NULL;

  label = gtk_list_item_get_child (list_item);
  row = (WordListModelRow *) gtk_list_item_get_item (list_item);
  text = g_strdup_printf ("%s — (%d)",
                          word_list_model_row_get_word (row),
                          word_list_model_row_get_priority (row));
  gtk_label_set_text (GTK_LABEL (label), text);
}

static void
edit_word_list_init (EditWordList *self)
{
  g_autoptr (GtkListItemFactory) across_factory = NULL;
  g_autoptr (GtkListItemFactory) down_factory = NULL;

  self->across_list_model = word_list_model_new ();
  self->down_list_model = word_list_model_new ();

  gtk_widget_init_template (GTK_WIDGET (self));

  across_factory = gtk_signal_list_item_factory_new ();
  g_signal_connect (across_factory, "setup", G_CALLBACK (setup_listitem_cb), NULL);
  g_signal_connect (across_factory, "bind", G_CALLBACK (bind_listitem_cb), NULL);

  down_factory = gtk_signal_list_item_factory_new ();
  g_signal_connect (down_factory, "setup", G_CALLBACK (setup_listitem_cb), NULL);
  g_signal_connect (down_factory, "bind", G_CALLBACK (bind_listitem_cb), NULL);

  g_object_set (self->across_word_list_view,
                "model", gtk_no_selection_new (G_LIST_MODEL (self->across_list_model)),
                "factory", across_factory,
                NULL);
  g_object_set (self->down_word_list_view,
                "model", gtk_no_selection_new (G_LIST_MODEL (self->down_list_model)),
                "factory", down_factory,
                NULL);

}

static void
edit_word_list_class_init (EditWordListClass *klass)
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->set_property = edit_word_list_set_property;
  object_class->get_property = edit_word_list_get_property;
  object_class->dispose = edit_word_list_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-word-list.ui");

  gtk_widget_class_bind_template_callback (widget_class, across_word_activate_cb);
  gtk_widget_class_bind_template_callback (widget_class, down_word_activate_cb);

  gtk_widget_class_bind_template_child (widget_class, EditWordList, across_word_list_view);
  gtk_widget_class_bind_template_child (widget_class, EditWordList, down_word_list_view);

  gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BIN_LAYOUT);

  obj_props[PROP_PUZZLE_STACK] = g_param_spec_object ("puzzle-stack",
                                                      NULL, NULL,
                                                      PUZZLE_TYPE_STACK,
                                                      G_PARAM_READWRITE);
  obj_props[PROP_EDIT_GRID] = g_param_spec_object ("edit-grid",
                                                   NULL, NULL,
                                                   EDIT_TYPE_GRID,
                                                   G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
  g_object_class_install_properties (object_class, N_PROPS, obj_props);
}

static void
edit_word_list_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
  EditWordList *self = EDIT_WORD_LIST (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_clear_object (&self->puzzle_stack);
      self->puzzle_stack = g_value_dup_object (value);
      if (self->puzzle_stack)
        {
          g_signal_connect_swapped (self->puzzle_stack,
                                    "changed",
                                    G_CALLBACK (puzzle_changed_cb),
                                    self);
          puzzle_changed_cb (self, TRUE, STACK_CHANGE_PUZZLE, self->puzzle_stack);
        }
      break;
    case PROP_EDIT_GRID:
      self->edit_grid = g_value_dup_object (value);
      g_signal_connect_swapped (self->edit_grid, "cursor-changed",
                                G_CALLBACK (grid_cursor_changed_cb),
                                self);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_word_list_get_property (GObject    *object,
                        guint       prop_id,
                        GValue     *value,
                        GParamSpec *pspec)
{
  EditWordList *self = EDIT_WORD_LIST (object);

  switch (prop_id)
    {
    case PROP_PUZZLE_STACK:
      g_value_set_object (value, self->puzzle_stack);
      break;
    case PROP_EDIT_GRID:
      g_value_set_object (value, self->edit_grid);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
edit_word_list_dispose (GObject *object)
{
  EditWordList *self;
  GtkWidget *child;

  self = EDIT_WORD_LIST (object);

  while ((child = gtk_widget_get_first_child (GTK_WIDGET (object))))
    gtk_widget_unparent (child);

  g_clear_object (&self->puzzle_stack);
  g_clear_object (&self->edit_grid);
  g_clear_pointer (&self->state, xword_state_free);

  G_OBJECT_CLASS (edit_word_list_parent_class)->dispose (object);
}

/* Called when a new frame is pushed on the stack */
static void
edit_word_list_puzzle_stack_changed_new (EditWordList *self,
                                         PuzzleStack  *puzzle_stack)
{
  IPuzPuzzle *puzzle;
  PuzzleStackChangeType change_type;

  puzzle = puzzle_stack_get_puzzle (puzzle_stack);
  change_type = puzzle_stack_get_change_type (puzzle_stack);

  if (puzzle == NULL)
    {
      g_clear_pointer (&self->state, xword_state_free);
      /* FIXME: set some sort of unloaded visual state */
    }
  else /* We have a puzzle to work with */
    {
      /* We always create a new state and reset the state when we get a new puzzle */
      if (change_type == STACK_CHANGE_PUZZLE)
        {
          CrosswordsQuirks *quirks;
          quirks = crosswords_quirks_new (puzzle);
          self->state = xword_state_replace (self->state, xword_state_new (IPUZ_CROSSWORD (puzzle), quirks, XWORD_STATE_EDIT));
        }
      else
        {
          /* We may have a similar looking puzzle, but it's a new-to-us puzzle
           * so we update the state as best as we can */
          if (self->state->xword != IPUZ_CROSSWORD (puzzle))
            {
              self->state = xword_state_replace (self->state, xword_state_dehydrate (self->state));
              self->state = xword_state_replace (self->state, xword_state_hydrate (self->state,
                                                                                   IPUZ_CROSSWORD (puzzle),
                                                                                   NULL));
            }
        }
    }

  if (self->state)
    {
      XwordState *dehydrated_state;
      dehydrated_state = xword_state_dehydrate (self->state);
      puzzle_stack_set_data (puzzle_stack,
                             "edit-word-list",
                             dehydrated_state,
                             (GDestroyNotify) xword_state_free);
    }
}

static void
edit_word_list_puzzle_stack_changed_load (EditWordList *self,
                                          PuzzleStack  *puzzle_stack)
{
  IPuzPuzzle *puzzle;
  XwordState *dehydrated_state;

  puzzle = puzzle_stack_get_puzzle (puzzle_stack);
  dehydrated_state = puzzle_stack_get_data (puzzle_stack, "edit-word-list");

  if (dehydrated_state)
    {
      self->state = xword_state_replace (self->state, xword_state_hydrate (dehydrated_state,
                                                                           IPUZ_CROSSWORD (puzzle),
                                                                           NULL));
    }
  else
    {
      if (self->state)
        g_clear_pointer (&self->state, xword_state_free);
    }
}

static void
puzzle_changed_cb (EditWordList          *self,
                   gboolean               new_frame,
                   PuzzleStackChangeType  change_type,
                   PuzzleStack           *puzzle_stack)
{
  IPuzPuzzle *puzzle;

  puzzle = puzzle_stack_get_puzzle (puzzle_stack);
  if (puzzle == NULL)
    return;

  if (new_frame)
    {
      edit_word_list_puzzle_stack_changed_new (self, puzzle_stack);
    }
  else
    {
      edit_word_list_puzzle_stack_changed_load (self, puzzle_stack);
    }


  /* Is this right? */
  if (change_type == STACK_CHANGE_PUZZLE ||
      change_type == STACK_CHANGE_GRID)
    edit_word_list_update_states (self);
}

static void
grid_cursor_changed_cb (EditWordList  *self,
                        IPuzCellCoord *cursor,
                        IPuzClueId    *clue_id)
{
  g_assert (EDIT_IS_WORD_LIST (self));

  /* this callback can be triggered by the edit_grid before we get our
   * puzzle set. It's harmless, but we need to check to make sure we
   * have our stack. */
  if (self->state == NULL)
    return;

  self->state->cursor = *cursor;
  self->state->clue = *clue_id;

  edit_word_list_update_state (self, self->state);
}

static void
word_activate (EditWordList         *self,
               guint              position,
               IPuzClueDirection  direction)
{
  GListModel *model;
  WordListModelRow *row;
  IPuzClue *clue;
  const GArray *coords;
  const gchar *ptr;
  guint i = 0;

  if (direction == IPUZ_CLUE_DIRECTION_ACROSS)
    model = G_LIST_MODEL (self->across_list_model);
  else
    model = G_LIST_MODEL (self->down_list_model);

  row = g_list_model_get_item (model, position);
  ptr = word_list_model_row_get_word (row);
  clue = ipuz_crossword_find_clue_by_coord (self->state->xword,
                                            direction,
                                            self->state->cursor);

  /* Some sanity checking. May fail in raw mode in the future */
  g_assert (ptr && *ptr);
  g_assert (clue != NULL);


  coords = ipuz_clue_get_cells (clue);

  while (ptr[0])
    {
      IPuzCellCoord coord;
      IPuzCell *cell;
      g_autofree gchar *substr = NULL;

      g_assert (i < coords->len);

      substr = g_utf8_substring (ptr, 0, 1);

      coord = g_array_index (coords, IPuzCellCoord, i);
      cell = ipuz_crossword_get_cell (IPUZ_CROSSWORD (self->state->xword), coord);

      ipuz_cell_set_solution (cell, substr);

      ptr = g_utf8_next_char (ptr);
      i++;
    }

  puzzle_stack_push_change (self->puzzle_stack, STACK_CHANGE_GRID, IPUZ_PUZZLE (self->state->xword));
}

static void
across_word_activate_cb (EditWordList *self,
                         guint      position)
{
  word_activate (self, position, IPUZ_CLUE_DIRECTION_ACROSS);
}

static void
down_word_activate_cb (EditWordList *self,
                       guint      position)
{
  word_activate (self, position, IPUZ_CLUE_DIRECTION_DOWN);
}

static void
edit_word_list_update_state (EditWordList *self,
                             XwordState   *state)
{
  IPuzClue *across_clue;
  IPuzClue *down_clue;
  g_autofree gchar *across_clue_string = NULL;
  g_autofree gchar *down_clue_string = NULL;

  if (! XWORD_STATE_HAS_CURSOR (state))
    {
      word_list_model_set_filter (self->across_list_model, NULL);
      word_list_model_set_filter (self->down_list_model, NULL);
      return;
    }


  across_clue = ipuz_crossword_find_clue_by_coord (state->xword,
                                                   IPUZ_CLUE_DIRECTION_ACROSS,
                                                   state->cursor);
  down_clue = ipuz_crossword_find_clue_by_coord (state->xword,
                                                 IPUZ_CLUE_DIRECTION_DOWN,
                                                 state->cursor);

  if (across_clue)
    {
      IPuzClueId clue_id = ipuz_crossword_get_clue_id (state->xword, across_clue);
      across_clue_string = ipuz_crossword_get_clue_string_by_id (state->xword, clue_id);
    }

  if (down_clue)
    {
      IPuzClueId clue_id = ipuz_crossword_get_clue_id (state->xword, down_clue);
      down_clue_string = ipuz_crossword_get_clue_string_by_id (state->xword, clue_id);
    }

  word_list_model_set_filter (self->across_list_model, across_clue_string);
  word_list_model_set_filter (self->down_list_model, down_clue_string);
}

static void
edit_word_list_update_states (EditWordList *self)
{
  if (self->state == NULL)
    return;

  edit_word_list_update_state (self, self->state);
}
